"""
Provides Base Classes for this project.
"""

import logging
from gzip import compress
from typing import Callable, Coroutine, Any

import httpx
from httpx import Response, HTTPError

from .exceptions import raise_error

logger = logging.getLogger("anaplan_sdk")


class _BaseClient:
    def __init__(self, retry_count: int, client: httpx.Client):
        self._retry_count = retry_count
        self._client = client

    def _get(self, url: str) -> dict[str, float | int | str | list | dict | bool]:
        return self._run_with_retry(self._client.get, url).json()

    def _get_binary(self, url: str) -> bytes:
        return self._run_with_retry(self._client.get, url).content

    def _post(self, url: str, json: dict) -> dict[str, float | int | str | list | dict | bool]:
        return self._run_with_retry(
            self._client.post, url, headers={"Content-Type": "application/json"}, json=json
        ).json()

    def _post_empty(self, url: str) -> None:
        self._run_with_retry(self._client.post, url)

    def _put_binary_gzip(self, url: str, content: bytes) -> Response:
        return self._run_with_retry(
            self._client.put,
            url,
            headers={"Content-Type": "application/x-gzip"},
            content=compress(content),
        )

    def _run_with_retry(self, func: Callable[..., Response], *args, **kwargs) -> Response:
        for i in range(max(self._retry_count, 1)):
            try:
                response = func(*args, **kwargs)
                response.raise_for_status()
                return response
            except HTTPError as error:
                url = args[0] or kwargs.get("url")
                if i >= self._retry_count - 1:
                    raise_error(error)
                logger.info(f"Retrying for: {url}")


class _AsyncBaseClient:
    def __init__(self, retry_count: int, client: httpx.AsyncClient):
        self._retry_count = retry_count
        self._client = client

    async def _get(self, url: str) -> dict[str, float | int | str | list | dict | bool]:
        return (await self._run_with_retry(self._client.get, url)).json()

    async def _get_binary(self, url: str) -> bytes:
        return (await self._run_with_retry(self._client.get, url)).content

    async def _post(
        self, url: str, json: dict
    ) -> dict[str, float | int | str | list | dict | bool]:
        return (
            await self._run_with_retry(
                self._client.post, url, headers={"Content-Type": "application/json"}, json=json
            )
        ).json()

    async def _post_empty(self, url: str) -> None:
        await self._run_with_retry(self._client.post, url)

    async def _put_binary_gzip(self, url: str, content: bytes) -> Response:
        return await self._run_with_retry(
            self._client.put,
            url,
            headers={"Content-Type": "application/x-gzip"},
            content=compress(content),
        )

    async def _run_with_retry(
        self, func: Callable[..., Coroutine[Any, Any, Response]], *args, **kwargs
    ) -> Response:
        for i in range(max(self._retry_count, 1)):
            try:
                response = await func(*args, **kwargs)
                response.raise_for_status()
                return response
            except HTTPError as error:
                if i >= self._retry_count - 1:
                    raise_error(error)



import httpx

from ._base import _BaseClient
from .models import ModelStatus, LineItem, List, ListMetadata, ListItem, InsertionResult, Module


class _TransactionalClient(_BaseClient):
    def __init__(self, client: httpx.Client, model_id: str, retry_count: int) -> None:
        self._client = client
        self._url = f"https://api.anaplan.com/2/0/models/{model_id}"
        super().__init__(retry_count, client)

    def list_modules(self) -> list[Module]:
        """
        Lists all the Modules in the Model.
        :return: All Modules on this model as a list of :py:class:`Module`.
        """
        return [Module.model_validate(e) for e in self._get(f"{self._url}/modules").get("modules")]

    def get_model_status(self) -> ModelStatus:
        """
        Gets the current status of the Model.
        :return: The current status of the Model as a :py:class:`ModelStatus`.
        """
        return ModelStatus.model_validate(self._get(f"{self._url}/status").get("requestStatus"))

    def list_line_items(self, only_module_id: int | None = None) -> list[LineItem]:
        """
        Lists all the Line Items in the Model.
        :param only_module_id: If provided, only Line Items from this Module will be returned.
        :return: All Line Items on this model as a list of :py:class:`LineItem`.
        """
        url = (
            f"{self._url}/modules/{only_module_id}/lineItems?includeAll=true"
            if only_module_id
            else f"{self._url}/lineItems?includeAll=true"
        )
        return [LineItem.model_validate(e) for e in self._get(url).get("items")]

    def list_lists(self) -> list[List]:
        """
        Lists all the Lists in the Model.
        :return: All Lists on this model as a list of :py:class:`List`.
        """
        return [List.model_validate(e) for e in self._get(f"{self._url}/lists").get("lists")]

    def get_list_metadata(self, list_id: int) -> ListMetadata:
        """
        Gets the metadata for a List.
        :param list_id: The ID of the List.
        :return: The metadata for the List as a :py:class:`ListMetadata`.
        """
        return ListMetadata.model_validate(
            self._get(f"{self._url}/lists/{list_id}").get("metadata")
        )

    def get_list_items(self, list_id: int) -> list[ListItem]:
        """
        Gets all the items in a List.
        :param list_id: The ID of the List.
        :return: All items in the List as a list of :py:class:`ListItem`.
        """
        return [
            ListItem.model_validate(e)
            for e in self._get(f"{self._url}/lists/{list_id}/items?includeAll=true").get(
                "listItems"
            )
        ]

    def add_items_to_list(
        self, list_id: int, items: list[dict[str, str | dict]]
    ) -> InsertionResult:
        """
        Adds items to a List.
        :param list_id: The ID of the List.
        :param items: The items to add to the List.
        :return: The result of the insertion as an :py:class:`InsertionResult`.
        """
        # TODO: Paginate by 100k records.
        return InsertionResult.model_validate(
            self._post(f"{self._url}/lists/{list_id}/items?action=add", json={"items": items})
        )

    def delete_list_items(self, list_id: int, items: list[dict[str, str | int]]) -> None:
        """
        Deletes items from a List.
        :param list_id: The ID of the List.
        :param items: The items to delete from the List. Must be a dict with either `code` or `id`
                      as the keys to identify the records to delete.
        """
        self._post(f"{self._url}/lists/{list_id}/items?action=delete", json={"items": items})

    def reset_list_index(self, list_id: int) -> None:
        """
        Resets the index of a List. The List must be empty to do so.
        :param list_id: The ID of the List.
        """
        self._post_empty(f"{self._url}/lists/{list_id}/resetIndex")



from typing import Literal

from pydantic import BaseModel, Field, field_validator

from .exceptions import InvalidIdentifierException


class Workspace(BaseModel):
    """
    Object representing an Anaplan Workspace.

    **id (str):** The unique identifier of this workspace.

    **name (str):** The name of this workspace that is also displayed to the users. This can
    change any time.

    **active (bool):** Whether this workspace is active or not.

    **size_allowance (int):** The maximum allowed size of this workspace in bytes.

    **current_size (int):** The current size of this workspace in bytes.
    """

    id: str
    name: str
    active: bool
    size_allowance: int = Field(alias="sizeAllowance", ge=0)
    current_size: int = Field(alias="currentSize", ge=0)


class Model(BaseModel):
    """
    Object representing an Anaplan Model.

    **id (str):** The unique identifier of this model.

    **name (str):** The name of this model that is also displayed to the users. This can
    change any time.

    **active_state (str):** The current state of this model. One of "ARCHIVED", "UNLOCKED", "ACTIVE"

    **last_saved_serial_number (int):** The serial number of the last save of this model.

    **last_modified_by_user_guid (str):** The unique identifier of the user who last modified this
    model.

    **memory_usage (int):** The memory usage of this model in bytes.

    **current_workspace_id (str):** The unique identifier of the workspace that this model is
    currently in.

    **current_workspace_name (str):** The name of the workspace that this model is currently in.

    **url (str):** The URL of this model.

    **category_values (list):** The category values of this model.

    **iso_creation_date (str):** The creation date of this model in ISO format.

    **last_modified (str):** The last modified date of this model in ISO format.
    """

    id: str
    name: str
    active_state: str = Field(alias="activeState")
    last_saved_serial_number: int = Field(alias="lastSavedSerialNumber")
    last_modified_by_user_guid: str = Field(alias="lastModifiedByUserGuid")
    memory_usage: int | None = Field(None, alias="memoryUsage")
    current_workspace_id: str = Field(alias="currentWorkspaceId")
    current_workspace_name: str = Field(alias="currentWorkspaceName")
    url: str = Field(alias="modelUrl")
    category_values: list = Field(alias="categoryValues")
    iso_creation_date: str = Field(alias="isoCreationDate")
    last_modified: str = Field(alias="lastModified")


class File(BaseModel):
    """
    Object representing an Anaplan File.

    **id (int):** The unique identifier of this file.

    **name (str):** The name of this file.

    **chunk_count (int):** The number of chunks this file is split into.

    **delimiter (str):** The delimiter used in this file.

    **encoding (str):** The encoding of this file.

    **first_data_row (int):** The row number of the first data row in this file.

    **format (str):** The format of this file.

    **header_row (int):** The row number of the header row in this file.

    **separator (str):** The separator used in this file.
    """

    id: int
    name: str
    chunk_count: int = Field(alias="chunkCount")
    delimiter: str | None = Field(None)
    encoding: str | None = Field(None)
    first_data_row: int = Field(alias="firstDataRow")
    format: str | None = Field(None)
    header_row: int = Field(alias="headerRow")
    separator: str | None = Field(None)


class List(BaseModel):
    """
    Object representing an Anaplan List.

    **id (int):** The unique identifier of this list.

    **name (str):** The name of this list.
    """

    id: int
    name: str


class ListMetadata(BaseModel):
    """
    Object representing the metadata of an Anaplan List.

    **id (int):** The unique identifier of this list.

    **name (str):** The name of this list.

    **has_selective_access (bool):** Whether this list has selective access or not.

    **properties (list):** The properties of this list.

    **production_data (bool):** Whether this list is production data or not.

    **managed_by (str):** The user who manages this list.

    **numbered_list (bool):** Whether this list is a numbered list or not.

    **use_top_level_as_page_default (bool):** Whether the top level is used as the page default
    or not.

    **item_count (int):** The number of items in this list.

    **next_item_index (int):** The index of the next item in this list.

    **workflow_enabled (bool):** Whether the workflow is enabled for this list or not.

    **permitted_items (int):** The number of permitted items in this list.

    **used_in_applies_to (str):** The applies to value of this list.
    """

    id: int
    name: str
    has_selective_access: bool = Field(alias="hasSelectiveAccess")
    properties: list = Field([])
    production_data: bool = Field(alias="productionData")
    managed_by: str = Field(alias="managedBy")
    numbered_list: bool = Field(alias="numberedList")
    use_top_level_as_page_default: bool = Field(alias="useTopLevelAsPageDefault")
    item_count: int = Field(alias="itemCount")
    next_item_index: int | None = Field(None, alias="nextItemIndex")
    workflow_enabled: bool = Field(alias="workflowEnabled")
    permitted_items: int = Field(alias="permittedItems")
    used_in_applies_to: str | None = Field(None, alias="usedInAppliesTo")


class ListItem(BaseModel):
    """
    Object representing an Anaplan List Item.

    **id (int):** The unique identifier of this list item.

    **name (str):** The name of this list item.

    **code (str):** The code of this list item.

    **properties (dict):** The properties of this list item.

    **subsets (dict):** The subsets of this list item.

    **parent (str):** The parent of this list item.

    **parent_id (str):** The unique identifier of the parent of this list item.
    """

    id: int
    name: str
    code: str | None = Field(None)
    properties: dict = Field({})
    subsets: dict = Field({})
    parent: str | None = Field(None)
    parent_id: str | None = Field(None, alias="parentId")


class Action(BaseModel):
    """
    Object representing an Anaplan Action.

    **id (int):** The unique identifier of this action.

    **name (str):** The name of this action.

    **type (str):** The type of this action.
    """

    id: int
    name: str
    type: str = Field(alias="actionType")


class Process(BaseModel):
    """
    Object representing an Anaplan Process.

    **id (int):** The unique identifier of this process.

    **name (str):** The name of this process.
    """

    id: int
    name: str


class Import(BaseModel):
    """
    Object representing an Anaplan Import.

    **id (int):** The unique identifier of this import.

    **name (str):** The name of this import.

    **type (str):** The type of this import.

    **source_id (int):** The unique identifier of the data source of this import.
    """

    id: int
    name: str
    type: str = Field(alias="importType")
    source_id: int | None = Field(None, alias="importDataSourceId")

    # noinspection PyNestedDecorators
    @field_validator("source_id", mode="before")
    @classmethod
    def _empty_source_is_none(cls, inp: str):
        return inp if inp else None


class Export(BaseModel):
    """
    Object representing an Anaplan Export.

    **id (int):** The unique identifier of this export.

    **name (str):** The name of this export.

    **type (str):** The type of this export.

    **format (str):** The format of this export.

    **encoding (str):** The encoding of this export.

    **layout (str):** The layout of this export.
    """

    id: int
    name: str
    type: str = Field(alias="exportType")
    format: str = Field(alias="exportFormat")
    encoding: str
    layout: str


class Module(BaseModel):
    """
    Object representing an Anaplan Module.

    **id (int):** The unique identifier of this module.

    **name (str):** The name of this module.
    """

    id: int
    name: str


class ModelStatus(BaseModel):
    """
    Object representing the status of an Anaplan Model.

    **peak_memory_usage_estimate (int):** The peak memory usage estimate of this model.

    **peak_memory_usage_time (int):** The peak memory usage time of this model.

    **progress (float):** The progress of this model.

    **current_step (str):** The current step of this model.

    **tooltip (str):** The tooltip of this model.

    **task_id (str):** The unique identifier of the task of this model.

    **creation_time (int):** The creation time of this model.

    **export_task_type (str):** The export task type of this model.
    """

    peak_memory_usage_estimate: int | None = Field(alias="peakMemoryUsageEstimate")
    peak_memory_usage_time: int | None = Field(alias="peakMemoryUsageTime")
    progress: float
    current_step: str = Field(alias="currentStep")
    tooltip: str | None
    task_id: str | None = Field(alias="taskId")
    creation_time: int = Field(alias="creationTime")
    export_task_type: str | None = Field(alias="exportTaskType")


class LineItem(BaseModel):
    """
    Object representing an Anaplan Line Item.

    **id (int):** The unique identifier of this line item.

    **name (str):** The name of this line item.

    **module_id (int):** The unique identifier of the module this line item belongs to.

    **module_name (str):** The name of the module this line item belongs to.

    **format (str):** The format of this line item.

    **format_metadata (dict):** The format metadata of this line item.

    **summary (str):** The summary of this line item.

    **applies_to (list):** The applies to value of this line item.

    **time_scale (str):** The time scale of this line item.

    **time_range (str):** The time range of this line item.

    **version (dict):** The version of this line item.

    **style (str):** The style of this line item.

    **cell_count (int):** The cell count of this line item.

    **notes (str):** The notes of this line item.

    **is_summary (bool):** Whether this line item is a summary or not.

    **formula (str):** The formula of this line item.

    **formula_scope (str):** The formula scope of this line item.

    **use_switchover (bool):** Whether the switchover is used or not.

    **breakback (bool):** Whether the breakback is enabled or not.

    **brought_forward (bool):** Whether the brought forward is enabled or not.

    **start_of_section (bool):** Whether this line item is the start of a section or not.
    """

    id: int
    name: str
    module_id: int = Field(alias="moduleId")
    module_name: str = Field(alias="moduleName")
    format: str
    format_metadata: dict = Field(alias="formatMetadata")
    summary: str
    applies_to: list[dict] = Field([], alias="appliesTo")
    time_scale: str = Field(alias="timeScale")
    time_range: str = Field(alias="timeRange")
    version: dict
    style: str
    cell_count: int | None = Field(None, alias="cellCount")
    notes: str
    is_summary: bool = Field(alias="isSummary")
    formula: str | None = Field(None)
    formula_scope: str = Field(alias="formulaScope")
    use_switchover: bool = Field(alias="useSwitchover")
    breakback: bool
    brought_forward: bool = Field(alias="broughtForward")
    start_of_section: bool = Field(alias="startOfSection")


class Failure(BaseModel):
    """
    Object representing a failure in an Anaplan import or export.

    **index (int):** The index of the item that failed.

    **reason (str):** The reason for the failure.

    **details (str):** The details of the failure.
    """

    index: int = Field(alias="requestIndex")
    reason: str = Field(alias="failureType")
    details: str = Field(alias="failureMessageDetails")


class InsertionResult(BaseModel):
    """
    Object representing the result of an insertion in an Anaplan list.

    **added (int):** The number of items successfully added.

    **ignored (int):** The number of items ignored, or items that failed.

    **total (int):** The total number of items.

    **failures (list):** The list of failures, see :py:class:`Failure`.
    """

    added: int
    ignored: int
    total: int
    failures: list[Failure] = Field([])


def action_url(action_id: int) -> Literal["imports", "exports", "actions", "processes"]:
    """
    Determine the type of action based on its identifier.
    :param action_id: The identifier of the action.
    :return: The type of action.
    """
    if 12000000000 <= action_id < 113000000000:
        return "imports"
    if 116000000000 <= action_id < 117000000000:
        return "exports"
    if 117000000000 <= action_id < 118000000000:
        return "actions"
    if 118000000000 <= action_id < 119000000000:
        return "processes"
    raise InvalidIdentifierException(f"Action '{action_id}' is not a valid identifier.")



"""
Synchronous Client.
"""

import logging
import time
from concurrent.futures import ThreadPoolExecutor

import httpx

from ._auth import AnaplanBasicAuth, AnaplanCertAuth, get_certificate, get_private_key
from ._base import _BaseClient
from ._transactional_client import _TransactionalClient
from .exceptions import (
    AnaplanActionError,
    InvalidIdentifierException,
)
from .models import Import, Export, Process, File, Action, Workspace, Model, action_url

logging.getLogger("httpx").setLevel(logging.CRITICAL)
logger = logging.getLogger("anaplan_sdk")


class Client(_BaseClient):
    """
    A synchronous Client for pythonic access to the Anaplan Integration API v2:
    https://anaplan.docs.apiary.io/. This Client provides high-level abstractions over the API, so
    you can deal with python objects and simple functions rather than implementation details like
    http, json, compression, chunking etc.


    For more information, quick start guides and detailed instructions refer to:
    https://vinzenzklass.github.io/anaplan-sdk.
    """

    def __init__(
        self,
        workspace_id: str | None = None,
        model_id: str | None = None,
        user_email: str | None = None,
        password: str | None = None,
        certificate: str | bytes | None = None,
        private_key: str | bytes | None = None,
        private_key_password: str | bytes | None = None,
        timeout: int = 30,
        retry_count: int = 2,
        status_poll_delay: int = 1,
        upload_parallel: bool = True,
        upload_chunk_size: int = 25_000_000,
        allow_file_creation: bool = False,
    ) -> None:
        """
        A synchronous Client for pythonic access to the Anaplan Integration API v2:
        https://anaplan.docs.apiary.io/. This Client provides high-level abstractions over the API,
        so you can deal with python objects and simple functions rather than implementation details
        like http, json, compression, chunking etc.


        For more information, quick start guides and detailed instructions refer to:
        https://vinzenzklass.github.io/anaplan-sdk.

        :param workspace_id: The Anaplan workspace Id. You can copy this from the browser URL or
                             find them using an HTTP Client like Postman, Paw, Insomnia etc.
        :param model_id: The identifier of the model.
        :param user_email: A valid email registered with the Anaplan Workspace you are attempting
                           to access. **The associated user must have Workspace Admin privileges**
        :param password: Password for the given `user_email`. This is not suitable for production
                         setups. If you intend to use this in production, acquire a client
                         certificate as described under: https://help.anaplan.com/procure-ca-certificates-47842267-2cb3-4e38-90bf-13b1632bcd44
        :param certificate: The absolute path to the client certificate file or the certificate
                            itself.
        :param private_key: The absolute path to the private key file or the private key itself.
        :param private_key_password: The password to access the private key if there is one.
        :param timeout: The timeout in seconds for the HTTP requests.
        :param retry_count: The number of times to retry an HTTP request if it fails. Set this to 0
                            to never retry. Defaults to 2, meaning each HTTP Operation will be
                            tried a total number of 2 times.
        :param status_poll_delay: The delay between polling the status of a task.
        :param upload_parallel: Whether to upload the chunks in parallel. Defaults to True. **If
                                you are heavily network bound or are experiencing rate limiting
                                issues, set this to False.**
        :param upload_chunk_size: The size of the chunks to upload. This is the maximum size of
                                  each chunk. Defaults to 25MB.
        :param allow_file_creation: Whether to allow the creation of new files. Defaults to False
                                    since this is typically unintentional and may well be unwanted
                                    behaviour in the API altogether. A file that is created this
                                    way will not be referenced by any action in anaplan until
                                    manually assigned so there is typically no value in dynamically
                                    creating new files and uploading content to them.
        """
        if not ((user_email and password) or (certificate and private_key)):
            raise ValueError(
                "Either `certificate` and `private_key` or `user_email` and `password` must be "
                "provided."
            )
        client = httpx.Client(
            auth=(
                AnaplanCertAuth(
                    get_certificate(certificate), get_private_key(private_key, private_key_password)
                )
                if certificate
                else AnaplanBasicAuth(user_email=user_email, password=password)
            ),
            timeout=timeout,
        )
        self._url = f"https://api.anaplan.com/2/0/workspaces/{workspace_id}/models/{model_id}"
        self._transactional_client = (
            _TransactionalClient(client, model_id, retry_count) if model_id else None
        )
        self.status_poll_delay = status_poll_delay
        self.upload_parallel = upload_parallel
        self.upload_chunk_size = upload_chunk_size
        self.allow_file_creation = allow_file_creation
        super().__init__(retry_count, client)

    @property
    def transactional(self) -> _TransactionalClient:
        """
        The Transactional Client provides access to the Anaplan Transactional API. This is useful
        for more advanced use cases where you need to interact with the Anaplan Model in a more
        granular way.

        If you instantiated the client without the field `model_id`, this will raise a
        :py:class:`ValueError`, since none of the endpoints can be invoked without the model Id.
        :return: The Transactional Client.
        """
        if not self._transactional_client:
            raise ValueError(
                "Cannot use the Transactional Client (Anaplan Transactional API) "
                "without field `model_id`. Make sure the instance you are trying to call this on "
                "is instantiated correctly with a valid `model_id`."
            )
        return self._transactional_client

    def list_workspaces(self) -> list[Workspace]:
        """
        Lists all the Workspaces the authenticated user has access to.
        :return: All Workspaces as a list of :py:class:`Workspace`.
        """
        return [
            Workspace.model_validate(e)
            for e in self._get("https://api.anaplan.com/2/0/workspaces?tenantDetails=true").get(
                "workspaces"
            )
        ]

    def list_models(self) -> list[Model]:
        """
        Lists all the Models the authenticated user has access to.
        :return: All Models in the Workspace as a list of :py:class:`Model`.
        """
        return [
            Model.model_validate(e)
            for e in self._get("https://api.anaplan.com/2/0/models?modelDetails=true").get("models")
        ]

    def list_files(self) -> list[File]:
        """
        Lists all the Files in the Model.
        :return: All Files on this model as a list of :py:class:`File`.
        """
        return [File.model_validate(e) for e in self._get(f"{self._url}/files").get("files")]

    def list_actions(self) -> list[Action]:
        """
        Lists all the Actions in the Model. This will only return the Actions listed under
        `Other Actions` in Anaplan. For Imports, exports, and processes, see their respective
        methods instead.
        :return: All Actions on this model as a list of :py:class:`Action`.
        """
        return [
            Action.model_validate(e) for e in (self._get(f"{self._url}/actions")).get("actions")
        ]

    def list_processes(self) -> list[Process]:
        """
        Lists all the Processes in the Model.
        :return: All Processes on this model as a list of :py:class:`Process`.
        """
        return [
            Process.model_validate(e)
            for e in (self._get(f"{self._url}/processes")).get("processes")
        ]

    def list_imports(self) -> list[Import]:
        """
        Lists all the Imports in the Model.
        :return: All Imports on this model as a list of :py:class:`Import`.
        """
        return [
            Import.model_validate(e) for e in (self._get(f"{self._url}/imports")).get("imports")
        ]

    def list_exports(self) -> list[Export]:
        """
        Lists all the Exports in the Model.
        :return: All Exports on this model as a list of :py:class:`Export`.
        """
        return [
            Export.model_validate(e) for e in (self._get(f"{self._url}/exports")).get("exports")
        ]

    def run_action(self, action_id: int) -> None:
        """
        Runs the specified Anaplan Action and validates the spawned task. If the Action fails or
        completes with errors, will raise an :py:class:`AnaplanActionError`. Failed Tasks are
        usually not something you can recover from at runtime and often require manual changes in
        Anaplan, i.e. updating the mapping of an Import or similar. So, for convenience, this will
        raise an Exception to handle - if you for e.g. think that one of the uploaded chunks may
        have been dropped and simply retrying with new data may help - and not return the task
        status information that needs to be handled by the caller.

        If you need more information or control, you can use `invoke_action()` and
        `get_task_status()`.
        :param action_id: The identifier of the Action to run. Can be any Anaplan Invokable;
                          Processes, Imports, Exports, Other Actions.
        """
        task_id = self.invoke_action(action_id)
        task_status = self.get_task_status(action_id, task_id)

        while "COMPLETE" not in task_status.get("taskState"):
            time.sleep(self.status_poll_delay)
            task_status = self.get_task_status(action_id, task_id)

        if task_status.get("taskState") == "COMPLETE" and not task_status.get("result").get(
            "successful"
        ):
            raise AnaplanActionError(f"Task '{task_id}' completed with errors.")

    def get_file(self, file_id: int) -> bytes:
        """
        Retrieves the content of the specified file.
        :param file_id: The identifier of the file to retrieve.
        :return: The content of the file.
        """
        return self._get_binary(f"{self._url}/files/{file_id}")

    def upload_file(self, file_id: int, content: str | bytes) -> None:
        """
        Uploads the content to the specified file. If `upload_parallel` is set to True on the
        instance you are invoking this from, will attempt to upload the chunks in parallel for
        better performance. If you are network bound or are experiencing rate limiting issues, set
        `upload_parallel` to False.

        :param file_id: The identifier of the file to upload to.
        :param content: The content to upload. **This Content will be compressed before uploading.
                        If you are passing the Input as bytes, pass it uncompressed to avoid
                        redundant work.**
        """
        if isinstance(content, str):
            content = content.encode()
        chunks = [
            content[i : i + self.upload_chunk_size]
            for i in range(0, len(content), self.upload_chunk_size)
        ]
        self._set_chunk_count(file_id, len(chunks))
        if self.upload_parallel:
            with ThreadPoolExecutor(max_workers=4) as executor:
                executor.map(
                    self._upload_chunk, (file_id,) * len(chunks), range(len(chunks)), chunks
                )
        else:
            for index, chunk in enumerate(chunks):
                self._upload_chunk(file_id, index, chunk)
        logger.info(f"Content loaded to  File '{file_id}'.")

    def get_task_status(
        self, action_id: int, task_id: str
    ) -> dict[str, float | int | str | list | dict | bool]:
        """
        Retrieves the status of the specified task.
        :param action_id: The identifier of the action that was invoked.
        :param task_id: The identifier of the spawned task.
        :return: The status of the task as returned by the API. For more information
                 see: https://anaplan.docs.apiary.io.
        """
        return self._get(f"{self._url}/{action_url(action_id)}/{action_id}/tasks/{task_id}").get(
            "task"
        )

    def invoke_action(self, action_id: int) -> str:
        """
        You may want to consider using `run_action()` instead.

        Invokes the specified Anaplan Action and returns the spawned Task identifier. This is
        useful if you want to handle the Task status yourself or if you want to run multiple
        Actions in parallel.
        :param action_id: The identifier of the Action to run. Can be any Anaplan Invokable.
        :return: The identifier of the spawned Task.
        """
        response = self._post(
            f"{self._url}/{action_url(action_id)}/{action_id}/tasks",
            json={"localeName": "en_US"},
        )
        task_id = response.get("task").get("taskId")
        logger.info(f"Invoked Action '{action_id}', spawned Task: '{task_id}'.")
        return task_id

    def _upload_chunk(self, file_id: int, index: int, chunk: bytes) -> None:
        self._put_binary_gzip(f"{self._url}/files/{file_id}/chunks/{index}", content=chunk)

    def _set_chunk_count(self, file_id: int, num_chunks: int) -> None:
        if not self.allow_file_creation and not (113000000000 <= file_id <= 113999999999):
            raise InvalidIdentifierException(
                f"File with Id {file_id} does not exist. If you want to dynamically create files "
                "to avoid this error, set `allow_file_creation=True` on the calling instance. "
                "Make sure you have understood the implications of this before doing so. "
            )
        response = self._post(f"{self._url}/files/{file_id}", json={"chunkCount": num_chunks})
        optionally_new_file = int(response.get("file").get("id"))
        if optionally_new_file != file_id:
            if self.allow_file_creation:
                logger.info(f"Created new file with name '{file_id}', Id is {optionally_new_file}.")
                return
            raise InvalidIdentifierException(
                f"File with Id {file_id} did not exist and was created in Anaplan. You may want to "
                f"ask a model builder to remove it. If you want to dynamically create files "
                "to avoid this error, set `allow_file_creation=True` on the calling instance. "
                "Make sure you have understood the implications of this before doing so."
            )
